from typing_extensions import TypeVar, Generic, Callable, Awaitable, Any, Protocol
from dataclasses import dataclass
from datetime import timedelta
from multiprocessing import Process
from haskellian import Tree, promise as P
from pipeteer.pipelines import Pipeline, Context
from pipeteer.queues import ReadQueue, WriteQueue, ListQueue
from pipeteer.util import param_type

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
Ctx = TypeVar('Ctx', bound=Context)
Artifact = TypeVar('Artifact')

class Stop(Exception):
  ...

class WorkflowContext(Protocol):
  async def call(self, pipe: Pipeline[A, B, Any, Any], x: A, /) -> B:
    ...

@dataclass
class WkfContext(WorkflowContext, Generic[Ctx]):
  ctx: Ctx
  prefix: tuple[str, ...]
  states: list
  key: str
  step: int = 0

  async def call(self, pipe: Pipeline[A, B, Ctx, Any], x: A, /) -> B:
    self.step += 1
    if self.step < len(self.states):
      return self.states[self.step]
    else:
      Qin = pipe.input(self.ctx, prefix=self.prefix)
      await Qin.push(self.key, x)
      raise Stop()

@dataclass
class Workflow(Pipeline[A, B, Ctx, Artifact | Callable[[], Process]], Generic[A, B, Ctx, Artifact]):
  pipelines: list[Pipeline]
  call: Callable[[A, WorkflowContext], Awaitable[B]]

  def states(self, ctx: Ctx, prefix: tuple[str, ...]) -> ListQueue:
    return ctx.backend.list_queue(prefix + (self.name, '_states'), self.type)

  async def rerun(self, key: str, ctx: Ctx, *, prefix: tuple[str, ...]):
    states = await self.states(ctx, prefix=prefix).read(key)
    wkf_ctx = WkfContext(ctx, prefix + (self.name,), states, key)
    return await self.call(states[0], wkf_ctx)

  def run(self, Qout: WriteQueue[B], ctx: Ctx, *, prefix: tuple[str, ...] = ()):
    
    Qin = self.input(ctx, prefix=prefix)
    Qstates = self.states(ctx, prefix=prefix)

    @P.run
    async def runner(Qin: ReadQueue[A], Qout: WriteQueue[B], Qstates: ListQueue, ctx: Ctx, prefix: tuple[str, ...]):
      ctx = ctx.prefix(prefix + (self.name,))
      while True:
        try:
          key, val = await Qin.wait_any()
          
          states = (await Qstates.safe_read(key) or []) + [val]
          wkf_ctx = WkfContext(ctx, prefix + (self.name,), states, key)
          try:
            out = await self.call(states[0], wkf_ctx)
            await Qstates.pop(key)
            await Qout.push(key, out)

          except Stop:
            await Qstates.append(key, val)

          await Qin.pop(key)

        except Exception as e:
          ctx.log('Error', e, level='ERROR')

    procs = dict(_root=lambda: Process(target=runner, args=(Qin, Qout, Qstates, ctx, prefix)))
    for pipe in self.pipelines:
      procs |= pipe.run(Qin, ctx, prefix=prefix + (self.name,)) # type: ignore

    return { self.name: procs }
  

def workflow(pipelines: list[Pipeline[Any, Any, Ctx, Artifact]], name: str | None = None):
  def decorator(fn: Callable[[A, WorkflowContext], Awaitable[B]]) -> Workflow[A, B, Ctx, Artifact]:
    return Workflow(
      type=param_type(fn),
      name=name or fn.__name__,
      pipelines=pipelines,
      call=fn,
    )
  return decorator


from typing_extensions import TypeVar, Generic, AsyncIterable
from abc import abstractmethod
import asyncio
from datetime import timedelta
from pipeteer.queues import InexistentItem, Transactional

A = TypeVar('A')

class ReadQueue(Transactional, Generic[A]):
  """A read/pop-only view of a `Queue`"""

  @abstractmethod
  async def pop(self, key: str, /):
    """Delete a specific item from the queue
    Throws `ReadError`"""

  async def wait_any(self, *, reserve: timedelta | None = None, poll_interval: timedelta = timedelta(seconds=1)) -> tuple[str, A]:
    """Read any item from the queue, waiting if necessary
    - `reserve`: reservation timeout. If not acknowledged within this time, the item is visible again
    - Throws `InfraError`
    """
    while not (res := await self.read_any(reserve=reserve)):
      await asyncio.sleep(poll_interval.total_seconds())
    return res

  async def read_any(self, *, reserve: timedelta | None = None) -> tuple[str, A] | None:
    """Read any item from the queue if not empty
    - `reserve`: reservation timeout. If not acknowledged within this time, the item is visible again
    - Throws `InfraError`
    """
    async for key, val in self.items(reserve=reserve, max=1):
      return key, val

  @abstractmethod
  async def read(self, key: str, /, *, reserve: timedelta | None = None) -> A:
    """Read a specific item from the queue
    - `reserve`: reservation timeout. If not acknowledged within this time, the item is visible again
    - Throws `ReadError`
    """
  
  async def safe_read(self, key: str, /, *, reserve: timedelta | None = None) -> A | None:
    """Read a specific item from the queue
    - `reserve`: reservation timeout. If not acknowledged within this time, the item is visible again
    - Throws `InfraError`
    """
    try:
      return await self.read(key, reserve=reserve)
    except InexistentItem:
      ...
  
  @abstractmethod
  def items(self, *, reserve: timedelta | None = None, max: int | None = None) -> AsyncIterable[tuple[str, A]]:
    """Iterate over the queue's items
    - `reserve`: reservation reserve for each iterated item. If not acknowledged within this time, items will become visible again
    - `max`: maximum number of items to iterate over (and reserve)
    - Throws `InfraError`
    """
  
  async def has(self, key: str, /, *, reserve: timedelta | None = None) -> bool:
    """Check if a specific item is in the queue
    - `reserve`: reservation timeout, after which the item is visible again
    - Throws `InfraError`
    """
    return await self.safe_read(key, reserve=reserve) is not None
    
  
  async def keys(self) -> AsyncIterable[str]:
    async for key, _ in self.items(reserve=None, max=None):
      yield key
  
  async def values(self) -> AsyncIterable[A]:
    async for _, val in self.items(reserve=None, max=None):
      yield val

class WriteQueue(Transactional, Generic[A]):
  """A write-only view of a `Queue`"""
  @abstractmethod
  async def push(self, key: str, value: A):
    """Push an item into the queue
    Throws `InfraError`"""


class Queue(ReadQueue[A], WriteQueue[A], Generic[A]):
  """A key-value, point-readable queue"""

class ListQueue(Queue[list[A]]):
  @abstractmethod
  async def append(self, key: str, value: A):
    ...


from .queues import ReadQueue, WriteQueue, Queue, ListQueue, Transaction, Transactional
from .backend import Backend
from .pipelines import Context, activity, task, workflow, WorkflowContext

__all__ = [
  'ReadQueue', 'WriteQueue', 'Queue', 'ListQueue', 'Transaction', 'Transactional',
  'Backend',
  'Context', 'activity', 'task', 'workflow', 'WorkflowContext',
]


from typing import Sequence, TypeVar
from dataclasses import dataclass, field
from functools import cached_property
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
import zmq
from zmq.asyncio import Context
from pipeteer.backend import Backend
from pipeteer.queues import ListQueue, Queue, SqlQueue, ListSqlQueue, ZeroMQueue

A = TypeVar('A')

@dataclass
class DefaultBackend(Backend):
  engine: AsyncEngine
  ctx: Context = field(default_factory=Context)
  zmq_port: int = 5555

  @cached_property
  def pub_sock(self):
    sock = self.ctx.socket(zmq.PUB)
    sock.bind(f'tcp://*:{self.zmq_port}')
    return sock
  
  @cached_property
  def sub_sock(self):
    sock = self.ctx.socket(zmq.SUB)
    sock.connect(f'tcp://localhost:{self.zmq_port}')
    sock.setsockopt_string(zmq.SUBSCRIBE, '')
    return sock

  @classmethod
  def sqlite(cls, path: str):
    return cls(create_async_engine(f'sqlite+aiosqlite:///{path}'))
  
  @classmethod
  def sql(cls, url: str):
    return cls(create_async_engine(url))

  @staticmethod
  def key(path: Sequence[str]) -> str:
    return '-'.join(path) or 'root'

  def queue(self, path: Sequence[str], type: type[A]) -> Queue[A]:
    key = self.key(path)
    queue = SqlQueue(type, self.engine, table=key)
    return ZeroMQueue(queue, topic=key)
    
  def list_queue(self, path: Sequence[str], type: type[A]) -> ListQueue[A]:
    return ListSqlQueue(list[type], self.engine, table=self.key(path))
  
  def output(self, type: type[A]) -> Queue[A]:
    return self.queue(('output',), type)


from typing_extensions import TypeVar, Generic, Callable
from dataclasses import dataclass
from haskellian import Tree
from pipeteer.pipelines import Pipeline, Context
from pipeteer.queues import ReadQueue, WriteQueue
from pipeteer.util import param_type, type_arg, num_params, Func2or3

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
Ctx = TypeVar('Ctx', bound=Context)
Artifact = TypeVar('Artifact')

@dataclass
class Task(Pipeline[A, B, Ctx, Artifact], Generic[A, B, Ctx, Artifact]):
  call: Callable[[ReadQueue[A], WriteQueue[B], Ctx], Tree[Artifact]]

  def run(self, Qout: WriteQueue[B], ctx: Ctx, /, *, prefix: tuple[str, ...] = ()) -> Tree[Artifact]:
    Qin = self.input(ctx, prefix=prefix)
    return self.call(Qin, Qout, ctx)
  
def task(name: str | None = None):
  def decorator(fn: Func2or3[ReadQueue[A], WriteQueue[B], Ctx, Tree[Artifact]]) -> Task[A, B, Ctx, Artifact]:
    return Task(
      name=name or fn.__name__,
      type=type_arg(param_type(fn)),
      call=fn if num_params(fn) == 3 else (lambda Qin, Qout, _: fn(Qin, Qout)) # type: ignore
    )
      
  return decorator


from .errors import InexistentItem, InfraError, QueueError, ReadError
from .spec import ReadQueue, WriteQueue, Queue, ListQueue
from .transactions import Transaction, Transactional
from .impl.sql import SqlQueue, ListSqlQueue
from .impl.zeromq import ReadZeroMQueue, WriteZeroMQueue, ZeroMQueue

__all__ = [
  'InexistentItem', 'InfraError', 'QueueError', 'ReadError',
  'ReadQueue', 'WriteQueue', 'Queue', 'ListQueue',
  'Transaction', 'Transactional',
  'SqlQueue', 'ListSqlQueue',
  'ReadZeroMQueue', 'WriteZeroMQueue', 'ZeroMQueue',
]


from typing_extensions import TypeVar, Generic, Callable, Awaitable
from dataclasses import dataclass
from datetime import timedelta
from multiprocessing import Process
from haskellian import Tree, promise as P
from pipeteer.pipelines import Pipeline, Context
from pipeteer.queues import ReadQueue, WriteQueue, Transaction
from pipeteer.util import param_type, num_params, Func1or2

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
Ctx = TypeVar('Ctx', bound=Context)
Artifact = Callable[[], Process]

@dataclass
class Activity(Pipeline[A, B, Ctx, Artifact], Generic[A, B, Ctx]):
  call: Callable[[A, Ctx], Awaitable[B]]
  reserve: timedelta | None = None

  def run(self, Qout: WriteQueue[B], ctx: Ctx, *, prefix: tuple[str, ...] = ()) -> Tree[Artifact]:
    Qin = self.input(ctx, prefix=prefix)
    ctx = ctx.prefix(prefix + (self.name,))

    @P.run
    async def runner(Qin: ReadQueue[A], Qout: WriteQueue[B]):
      while True:
        try:
          k, x = await Qin.wait_any(reserve=self.reserve)
          y = await self.call(x, ctx)
          async with Transaction(Qin, Qout, autocommit=True):
            await Qout.push(k, y)
            await Qin.pop(k)

        except Exception as e:
          ctx.log(f'Error: {e}', level='ERROR')

    return { self.name: lambda: Process(target=runner, args=(Qin, Qout)) }

def activity(name: str | None = None, *, reserve: timedelta | None = timedelta(minutes=2)):
  def decorator(fn: Func1or2[A, Ctx, Awaitable[B]]) -> Pipeline[A, B, Ctx, Artifact]:
    return Activity(
      type=param_type(fn), reserve=reserve, name=name or fn.__name__,
      call=fn if num_params(fn) == 2 else (lambda x, _: fn(x)) # type: ignore
    )
      
  return decorator


from .pipeline import Pipeline, Context
from ._activity import activity, Activity
from ._task import task, Task
from ._workflow import workflow, Workflow, WorkflowContext

__all__ = [
  'Pipeline', 'Context',
  'activity', 'Activity',
  'task', 'Task',
  'workflow', 'Workflow', 'WorkflowContext'
]


from typing import TypeVar, Sequence
from abc import ABC, abstractmethod
from pipeteer.queues import Queue, ListQueue

A = TypeVar('A')
B = TypeVar('B')

class Backend(ABC):

  @abstractmethod
  def queue(self, path: Sequence[str], type: type[A]) -> Queue[A]:
    ...

  @abstractmethod
  def list_queue(self, path: Sequence[str], type: type[A]) -> ListQueue[A]:
    ...

  @abstractmethod
  def output(self, type: type[A]) -> Queue[A]:
    ...

  @classmethod
  def sqlite(cls, path: str):
    from .default import DefaultBackend
    return DefaultBackend.sqlite(path)
  
  @classmethod
  def sql(cls, url: str):
    from .default import DefaultBackend
    return DefaultBackend.sql(url)


from typing_extensions import TypeVar, Generic, Iterable, Coroutine, Any
import asyncio
from functools import cached_property
from dataclasses import dataclass, field
from datetime import timedelta
import zmq
from zmq.asyncio import Context
from pipeteer import Queue, WriteQueue, ReadQueue

T = TypeVar('T')

async def race(coros: Iterable[Coroutine[Any, Any, T]]) -> T:
  tasks = [asyncio.create_task(c) for c in coros]
  done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
  for task in pending:
      task.cancel()
  return done.pop().result()

@dataclass
class WriteZeroMQueue(WriteQueue[T], Generic[T]):
  queue: WriteQueue[T]
  topic: str
  url: str = 'tcp://*:5555'
  ctx: Context = field(default_factory=Context)

  @cached_property
  def pub(self):
    sock = self.ctx.socket(zmq.PUB)
    sock.connect(self.url)
    return sock

  async def push(self, key: str, value: T):
    res = await self.queue.push(key, value)
    await self.pub.send_string(self.topic)
    return res
  
@dataclass
class ReadZeroMQueue(ReadQueue[T], Generic[T]):
  queue: ReadQueue[T]
  topic: str
  url: str = 'tcp://localhost:5555'
  ctx: Context = field(default_factory=Context)

  @cached_property
  def sub(self):
    sock = self.ctx.socket(zmq.SUB)
    sock.connect(self.url)
    sock.setsockopt_string(zmq.SUBSCRIBE, self.topic)
    return sock

  async def wait_zmq(self):
    await self.sub.recv_string()

  async def exp_backoff(self, *, reserve: timedelta | None = None, poll_interval: timedelta = timedelta(seconds=1)) -> tuple[str, T]:
    t0 = poll_interval.total_seconds()
    exp = 1
    while True:
      result = await self.queue.wait_any(reserve=reserve)
      if result:
        return result
      
      await asyncio.sleep(t0 * exp)
      exp *= 2


  async def wait_any(self, *, reserve: timedelta | None = None, poll_interval: timedelta = timedelta(seconds=1)) -> tuple[str, T]:
    while True:
      first = await race([self.exp_backoff(reserve=reserve, poll_interval=poll_interval), self.wait_zmq()])
      if first:
        return first
      await self.wait_zmq()
      
      result = await self.queue.wait_any(reserve=reserve)
      if result:
        return result
      
  def pop(self, key: str):
    return self.queue.pop(key)
  
  def read(self, key: str, /, *, reserve: timedelta | None = None):
    return self.queue.read(key, reserve=reserve)
  
  def read_any(self, *, reserve: timedelta | None = None):
    return self.queue.read_any(reserve=reserve)
  
  def items(self, *, reserve: timedelta | None = None, max: int | None = None):
    return self.queue.items(reserve=reserve, max=max)
  
  def keys(self):
    return self.queue.keys()
  
  def has(self, key: str, /, *, reserve: timedelta | None = None):
    return self.queue.has(key, reserve=reserve)
  
  def values(self):
    return self.queue.values()
        
@dataclass
class ZeroMQueue(Queue[T], ReadZeroMQueue[T], WriteZeroMQueue[T], Generic[T]):
  queue: Queue[T] # type: ignore


from typing_extensions import TypeVar, Generic, Callable, Self, Protocol, Sequence
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, replace
from multiprocessing import Process
from haskellian import Tree, trees
from dslog import Logger
from pipeteer import Queue, WriteQueue, Backend

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D', contravariant=True)
Artifact = TypeVar('Artifact', covariant=True)

@dataclass
class Context:
  backend: Backend
  log: Logger = field(default_factory=Logger.click)

  def prefix(self, path: tuple[str, ...]) -> Self:
    key = '/'.join(path) or 'root'
    return replace(self, log=self.log.prefix(f'[{key}]'))

  @classmethod
  def sqlite(cls, path: str):
    return cls(Backend.sqlite(path))
  
  @classmethod
  def sql(cls, url: str):
    return cls(Backend.sql(url))

Ctx = TypeVar('Ctx', bound=Context)

class Executor(Protocol, Generic[D]):
  def __call__(self, path: Sequence[str], artifact: D, /) -> Process:
    ...

def default_executor(_, artifact: Callable[[], Process]) -> Process:
  return artifact()

@dataclass
class Pipeline(ABC, Generic[A, B, Ctx, Artifact]):
  type: type[A]
  name: str

  def input(self, ctx: Ctx, *, prefix: tuple[str, ...] = ()) -> Queue[A]:
    return ctx.backend.queue(prefix + (self.name,), self.type)

  @abstractmethod
  def run(self, Qout: WriteQueue[B], ctx: Ctx, /, *, prefix: tuple[str, ...] = ()) -> Tree[Artifact]:
    ...

  def run_all(
    self, Qout: WriteQueue[B], ctx: Ctx, *,
    prefix: tuple[str, ...] = (), executor: Executor[Artifact] = default_executor
  ):
    procs = self.run(Qout, ctx, prefix=prefix)
    procs = trees.path_map(procs, executor)
    
    for path, proc in trees.flatten(procs):
      key = '/'.join((k for k in path if k != '_root'))
      ctx.log(f'[{key}] Starting...')
      proc.start()
    
    for path, proc in trees.flatten(procs):
      key = '/'.join((k for k in path if k != '_root'))
      proc.join()
      ctx.log(f'[{key}] Stopping...')


from typing_extensions import TypeVar, Callable, Generic, get_type_hints, get_args
from pipeteer.queues import ReadQueue

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')

def param_type(fn: Callable, idx=0):
  return list(get_type_hints(fn).values())[idx]

def type_arg(generic: type, idx=0) -> type:
  return get_args(generic)[idx]

def num_params(fn) -> int:
  from inspect import signature
  return len(signature(fn).parameters)

Func1or2 = Callable[[A, B], C] | Callable[[A], C]
Func2or3 = Callable[[A, B, C], D] | Callable[[A, B], D]